//////////////////////////////////////////////////////////////////////
//
// ClipboardToFile: Copies contents of clipboard to a file
//
// (C) 2006 Claus Brod
// http://www.clausbrod.de
// Software blog: http://www.clausbrod.de/Blog/BlogOnSoftware
// 
// Usage: ClipboardToFile [fileprefix]
//
//////////////////////////////////////////////////////////////////////

#define _AFXDLL
#include <afxwin.h>
#include <stdio.h>
#include <tchar.h>

#include "clipids.h"

////////// Bitmap handling ///////////////////////////////////////////

// BitmapToFile - convert device-dependent bitmap to DIB,
// then write to file
bool BitmapToFile(CBitmap& bitmap, LPTSTR filename, CPalette* pPal = 0)
{
  ASSERT(bitmap.GetSafeHandle());

  // Use default palette unless a palette has been specified
  HPALETTE hPal;
  if (!pPal || ((hPal = (HPALETTE)pPal->GetSafeHandle())==NULL))
    hPal = (HPALETTE)GetStockObject(DEFAULT_PALETTE);

  // Get bitmap information
  BITMAP bm;
  bitmap.GetObject(sizeof(bm), &bm);

  // Calculate validate bits/pixel
  WORD biBits = bm.bmPlanes * bm.bmBitsPixel;
  if (biBits <= 1)
    biBits = 1;
  else if (biBits <= 4)
    biBits = 4;
  else if (biBits <= 8)
    biBits = 8;
  else
    biBits = 24;

  BITMAPINFOHEADER bi;
  ::memset(&bi, 0, sizeof(bi));
  bi.biSize		     = sizeof(BITMAPINFOHEADER);
  bi.biWidth		   = bm.bmWidth;
  bi.biHeight 		 = bm.bmHeight;
  bi.biPlanes 		 = 1;
  bi.biBitCount		 = biBits;
  bi.biCompression = BI_RGB;

  // Allocate memory for header and color table
  int nColors = (1 << bi.biBitCount);
  if( nColors > 256 )
    nColors = 0;   // no color table?
  DWORD	dwLen = sizeof(BITMAPINFOHEADER) + nColors * sizeof(RGBQUAD);
  HANDLE hDIB = GlobalAlloc(GMEM_FIXED, dwLen);
  if (!hDIB)
    return false;

  // Select the provided palette
  HDC hDC = GetDC(NULL);
  hPal = SelectPalette(hDC,hPal,FALSE);
  RealizePalette(hDC);

  // Inquire biSizeImage using GetDIBits
  LPBITMAPINFOHEADER lpbi = (LPBITMAPINFOHEADER)hDIB;
  *lpbi = bi;
  GetDIBits(hDC, (HBITMAP)bitmap.GetSafeHandle(), 0, bi.biHeight,
    NULL, (LPBITMAPINFO)lpbi, DIB_RGB_COLORS);
  bi = *lpbi;

  if (bi.biSizeImage == 0) {
    // Oops, we need to calculate biSizeImage ourselves.
    // Image scan lines are 32-bit aligned.
    bi.biSizeImage = ((((bm.bmWidth * biBits) + 31) & ~31) / 8) * bm.bmHeight;
  }

  bool gotBits = false;
  // Reallocate buffer in full size
  dwLen += bi.biSizeImage;
  HANDLE handle = GlobalReAlloc(hDIB, dwLen, GMEM_MOVEABLE);
  if (handle) {
    hDIB = handle;
    gotBits = (FALSE != GetDIBits(hDC, (HBITMAP)bitmap.GetSafeHandle(),
      0, bi.biHeight,
      (LPBYTE)hDIB + (bi.biSize + nColors * sizeof(RGBQUAD)),
      (LPBITMAPINFO)hDIB,	DIB_RGB_COLORS));
  }
  
  // Reselect the original palette
  SelectPalette(hDC,hPal,FALSE);
  ReleaseDC(NULL,hDC);

  if (!gotBits) {
    GlobalFree(hDIB);
    return false;
  }

  FILE *f;
  if (!_tfopen_s(&f, filename, _T("wb"))) {
    BITMAPFILEHEADER hdr;
    hdr.bfType = ((WORD)('M'<<8) | 'B');
    hdr.bfSize = (DWORD)(GlobalSize(hDIB) + sizeof(hdr));
    hdr.bfReserved1 = hdr.bfReserved2 = 0;

    // calculate offset
    LPBITMAPINFOHEADER lpbi = (LPBITMAPINFOHEADER)hDIB;
    hdr.bfOffBits = (DWORD)(sizeof(hdr) + lpbi->biSize);
    if (lpbi->biBitCount > 0 && lpbi->biBitCount <= 8) {
      hdr.bfOffBits += (1 << lpbi->biBitCount) * sizeof(RGBQUAD);
    } else {
      if (lpbi->biBitCount == 16 || lpbi->biBitCount == 32) {
        if (lpbi->biCompression == BI_BITFIELDS) {
          hdr.bfOffBits += 3*sizeof(DWORD);
        }
      } else if (lpbi->biBitCount == 24) {
        hdr.bfOffBits += lpbi->biClrUsed*sizeof(RGBQUAD);
      }
    }

    // Write file header, then DIB header and bitmap data
    ::fwrite(&hdr, 1, sizeof(hdr), f);
    ::fwrite(lpbi, 1, GlobalSize(hDIB), f);
    fclose(f);
  }
  GlobalFree(hDIB);
  return f != 0;
}

////////// Clipboard enumeration /////////////////////////////////////

bool ClipboardToFile(TCHAR *filename, UINT clipID)
{
  HANDLE hClip = GetClipboardData(clipID);
  if (!hClip) {
    return false;
  }

  bool ret = false;

  switch(clipID)
  {
  case CF_ENHMETAFILE:
  case CF_DSPENHMETAFILE:
    {
      HENHMETAFILE meta = ::CopyEnhMetaFile((HENHMETAFILE)hClip, filename);
      if (meta) {
        ::DeleteEnhMetaFile(meta);
        ret = true;
      }
    }
    break;

  case CF_METAFILEPICT:
  case CF_DSPMETAFILEPICT:
    {
      METAFILEPICT *hClipLocked = (METAFILEPICT *)GlobalLock(hClip);
      HMETAFILE meta = ::CopyMetaFile(hClipLocked->hMF, filename);
      if (meta) {
        ::DeleteMetaFile(meta);
        ret = true;
      }
    }
    break;

  case CF_BITMAP:
  case CF_DSPBITMAP:
    ret = BitmapToFile(*CBitmap::FromHandle((HBITMAP)hClip), filename);
    break;

  case CF_PALETTE:
    {  // special-case code for CF_PALETTE - still need testcases for this
      UINT numEntries = GetPaletteEntries((HPALETTE)hClip, 0, 0, 0);
      if (numEntries) {
        PALETTEENTRY *palette = (PALETTEENTRY*)malloc(sizeof(PALETTEENTRY)*numEntries);
        if (GetPaletteEntries((HPALETTE)hClip, 0, numEntries, palette)) {
          FILE *f;
          if (_tfopen_s(&f, filename, _T("wb"))) {
            _ftprintf(stderr, _T("ERROR: Cannot open %s\n"), filename);
          } else {
            ::fwrite(&numEntries, 1, sizeof(numEntries), f);
            ::fwrite(palette, 1, sizeof(PALETTEENTRY)*numEntries, f);
            fclose(f);
            ret = true;
          }
        }
        free(palette);
      }
    }
    break;

  case CF_HDROP:
  case CF_PENDATA:
  case CF_DSPTEXT:
    // We may need special-case handling for these formats as well.
    // However, I have not found testcases yet, so for now
    // we're simply falling through to the default method.
    _ftprintf(stderr, _T("ERROR: Unsupported clipboard format %d\n"), clipID);

  default:
    {
      if (clipID >= CF_GDIOBJFIRST && clipID <= CF_GDIOBJLAST) {
        _ftprintf(stderr, _T("ERROR: Unsupported clipboard format %d\n"), clipID);
        // Did not find any test cases, so we don't handle these
        // IDs explicitly yet; however, we fall through to the default
        // handler, hoping that it might happen to work even for CF_GDIOBJ*.
      }
      FILE *f;
      if (_tfopen_s(&f, filename, _T("wb"))) {
        _ftprintf(stderr, _T("ERROR: Cannot open %s\n"), filename);
      } else {
        void *pData = (void*)GlobalLock(hClip);
        if (pData) {
          SIZE_T sz = ::GlobalSize(pData);
          if (sz) {
            size_t written = fwrite(pData, 1, sz, f);
            ret = (written == sz);
          }
        }
        ::GlobalUnlock(hClip);
        fclose(f);
      }
    }
    break;
  }

  return ret;
}

bool EnumClipboard(const TCHAR *filename)
{
  bool ret = true;

  HWND owner = ::CreateWindow(_T("STATIC"), _T("static"), 
      0, 0, 0, 100, 100, 0, 0, 0, 0);
  ::OpenClipboard(owner);

  // prepare filename buffer
  TCHAR namebuf[MAX_PATH+64];
  _tcscpy_s(namebuf, _countof(namebuf), filename);
  size_t buflen = _tcslen(namebuf);
  namebuf[buflen] = 0;

  UINT nextFormat = ::EnumClipboardFormats(0);
  while (nextFormat) {
    // build full filename
    TCHAR *clipformatName = const_cast<TCHAR*>(FormatFromClipID(nextFormat));
    if (clipformatName) {
      // check if clipformat name contains 'dangerous' characters
      TCHAR *p = _tcspbrk(clipformatName, __T("/\\'\""));
      while (p) {
        *p = '_';
        p = _tcspbrk(clipformatName, __T("/\\'\""));
      }
      _tcscpy_s(namebuf+buflen, _countof(namebuf)-buflen, clipformatName);
    } else {
      _ltot_s(nextFormat, namebuf+buflen, _countof(namebuf)-buflen, 10);
    }

    if (!ClipboardToFile(namebuf, nextFormat)) {
      _ftprintf(stderr, _T("ERROR: Cannot write data to %s\n"), namebuf);
      ret = false;
    } else {
      _tprintf(_T("Clipboard format %d successfully written to %s.\n"), nextFormat, namebuf);
    }
    nextFormat = ::EnumClipboardFormats(nextFormat);
  }
  ::CloseClipboard();
  return ret;
}


int _tmain(int argc, _TCHAR* argv[])
{
  _tprintf(_T("ClipboardToFile - (C) 2006 Claus Brod, http://www.clausbrod.de\n\n"));

  if (argc != 2 && argc != 1) {
    _ftprintf(stderr, _T("Usage: %s [filenameprefix]\n"), argv[0]);
    return 2;
  }

  const TCHAR *fnameprefix = (argc > 1) ? argv[1] : _T("");
  ::OleInitialize(0);
  bool ret = EnumClipboard(fnameprefix);
  ::OleUninitialize();
  if (!ret) {
    _ftprintf(stderr, _T("ERROR: Failure to enumerate clipboard and store all formats.\n"));
    return 1;
  }

  return 0;
}
