//////////////////////////////////////////////////////////////////////
//
// FileToClipboard: Copies files in specified formats to
// the Windows clipboard
//
// (C) 2006 Claus Brod
// http://www.clausbrod.de
// Software blog: http://www.clausbrod.de/Blog/BlogOnSoftware
// 
// Usage: FileToClipboard file format
//
// Examples:
//    FileToClipboard foo.bmp CF_BITMAP
//       Uses predefined clipboard format for bitmaps
//
//    FileToClipboard funkyfile.funky CF_FUNKY
//       Registers a new clipboard format called CF_FUNKY
//
//    FileToClipboard myprivateida.ho 50222
//       Numerical specification of clipboard format
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <stdio.h>
#include <tchar.h>

#include "clipids.h"

HANDLE ReadFileToMemory(const TCHAR *filename)
{
  FILE *f;
  errno_t error = _tfopen_s(&f, filename, _T("rb"));
  if (error) {
    _ftprintf(stderr, _T("ERROR: Cannot open %s\n"), filename);
    return 0;
  }

  // get size of file
  fseek(f, 0, SEEK_END);
  long size=ftell(f);
  fseek(f, 0, SEEK_SET);

  // allocate memory
  HANDLE hMem = ::GlobalAlloc(GMEM_MOVEABLE, size);
  if (!hMem) {
    fclose(f);
    return 0;
  }

  LPVOID mem = ::GlobalLock(hMem);
  if (!mem) {
    fclose(f);
    ::GlobalFree(hMem);
    return 0;
  }

  // read the file into memory
  size_t bytes_read = fread(mem, 1, size, f);
  fclose(f);
  ::GlobalUnlock(mem);
  if (bytes_read != size) {
    ::GlobalFree(hMem);
    return 0;
  }

  return hMem;
}

HANDLE FileToMetafile(const TCHAR *filename)
{
  HMETAFILE mf = ::GetMetaFile(filename);
  HANDLE hMfpict = GlobalAlloc(GMEM_MOVEABLE, sizeof(METAFILEPICT));
  METAFILEPICT *mfpict = (METAFILEPICT *)GlobalLock(hMfpict);
  mfpict->hMF = mf;
  mfpict->mm = MM_ANISOTROPIC;
  mfpict->xExt = 0;
  mfpict->yExt = 0;
  _ftprintf(stderr, _T("TBD: Fix METAFILEPICT structure.\n"));
  GlobalUnlock(mfpict);
  return hMfpict;
}

HANDLE FileToPalette(const TCHAR *filename)
{
  HANDLE hpal = NULL;

  // still looking for testcases for this
  FILE *f;
  errno_t error = _tfopen_s(&f, filename, _T("rb"));
  if (error) {
    _ftprintf(stderr, _T("ERROR: Cannot open %s\n"), filename);
  } else {
    UINT numEntries = 0;
    ::fread(&numEntries, sizeof(numEntries), 1, f);
    fclose(f);
    if (numEntries) {
      LOGPALETTE *logpalette = 
        (LOGPALETTE*)malloc(sizeof(LOGPALETTE)+ sizeof(PALETTEENTRY)*numEntries);
      logpalette->palNumEntries = numEntries;
      logpalette->palVersion = 0x0300;
      ::fread(logpalette->palPalEntry, sizeof(PALETTEENTRY), numEntries, f);
      // get PALETTE handle
      HPALETTE hpal = CreatePalette(logpalette);
      free(logpalette);
    }
  }

  return hpal;
}

bool FileToClipboard(TCHAR *filename, TCHAR *format)
{
  UINT clipid = ClipIDFromFormat(format);
  if (!clipid)
    return false;

  // we need a dummy owner, or else SetClipboardData() will fail
  HWND owner = ::CreateWindow(_T("STATIC"), 
    _T("static"), 0, 0, 0, 100, 100, 0, 0, 0, 0);
  if (!owner)
    return false;

  if (!::OpenClipboard(owner)) {
    ::DestroyWindow(owner);
    return false;
  }

  HANDLE hClip = 0;
  bool hClipIsMemory = true;

  switch(clipid) 
  {
  case CF_BITMAP:
  case CF_DSPBITMAP:
    hClip = ::LoadImage(0, filename, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
    break;

  case CF_METAFILEPICT:
  case CF_DSPMETAFILEPICT:
    hClip = FileToMetafile(filename);
    break;

  case CF_ENHMETAFILE:
  case CF_DSPENHMETAFILE:
    hClip = (HANDLE)::GetEnhMetaFile(filename);
    hClipIsMemory = false;
    break;


  case CF_PALETTE:
    hClip = FileToPalette(filename);
    break;

  case CF_PENDATA:
  case CF_HDROP:
  case CF_DSPTEXT:
    _ftprintf(stderr, _T("%s not supported yet.\n"), FormatFromClipID(clipid));
    // fall-through

  default:
    hClip = ReadFileToMemory(filename);
    break;
  }

  if (!hClip) {
    _ftprintf(stderr, _T("ERROR: Cannot prepare data for clipboard (error code %d).\n"), GetLastError());
    ::DestroyWindow(owner);
    return false;
  }

  ::EmptyClipboard();
  HANDLE h = ::SetClipboardData(clipid, hClip);
  ::CloseClipboard();
  ::DestroyWindow(owner);

  return h != 0;
}

int _tmain(int argc, TCHAR *argv[])
{
  _tprintf(_T("FileToClipboard - (C) 2006 Claus Brod, http://www.clausbrod.de\n\n"));

  if (argc != 3) {
    _ftprintf(stderr, _T("Usage: %s file clipformat\n"), argv[0]);
    return 2;
  }

  if (!FileToClipboard(argv[1], argv[2])) {
    _ftprintf(stderr, _T("ERROR: Cannot write to clipboard.\n"));
    return 1;
  }

  return 0;
}
